"""Netbox module

This module contains the methods to connect with a Netbox REST server
and retrieve the devices inventory

Classes:
    Netbox: this class dinamically retrieve the inventory from Netbox
"""
# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument

import asyncio
import logging
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

from pydantic import BaseModel, validator, Field

import aiohttp
from suzieq.poller.controller.inventory_async_plugin import \
    InventoryAsyncPlugin
from suzieq.poller.controller.source.base_source import Source, SourceModel
from suzieq.shared.utils import get_sensitive_data
from suzieq.shared.exceptions import InventorySourceError, SensitiveLoadError

_DEFAULT_PORTS = {'http': 80, 'https': 443}

logger = logging.getLogger(__name__)


class NetboxServerModel(BaseModel):
    """Model containing data to connect with Netbox server
    """
    host: str
    protocol: str
    port: str

    class Config:
        """pydantic configuration
        """
        extra = 'forbid'


class NetboxSourceModel(SourceModel):
    """Netbox source validation model
    """
    tag: Optional[Any] = Field(default=['suzieq'])
    period: Optional[int] = Field(default=3600)
    token: str
    ssl_verify: Optional[bool] = Field(alias='ssl-verify')
    server: Union[str, NetboxServerModel] = Field(alias='url')
    run_once: Optional[bool] = Field(default=False, alias='run_once')

    @validator('server', pre=True)
    def validate_and_set(cls, url, values):
        """Validate the field 'url' and set the correct parameters
        """
        if isinstance(url, str):
            url_data = urlparse(url)
            host = url_data.hostname
            if not host:
                raise ValueError(f'Unable to parse hostname {url}')
            protocol = url_data.scheme or 'http'
            if protocol not in _DEFAULT_PORTS:
                raise ValueError(f'Unknown protocol {protocol}')
            port = url_data.port or _DEFAULT_PORTS.get(protocol)
            if not port:
                raise ValueError(f'Unable to parse port {url}')
            server = NetboxServerModel(host=host, port=port, protocol=protocol)
            ssl_verify = values['ssl_verify']
            if ssl_verify is None:
                if server.protocol == 'http':
                    ssl_verify = False
                else:
                    ssl_verify = True
            else:
                if server.protocol == 'http' and ssl_verify:
                    raise ValueError(
                        'Cannot use ssl_verify=True with http host')
            values['ssl_verify'] = ssl_verify
            return server
        elif isinstance(url, NetboxServerModel):
            return url
        else:
            raise ValueError('Unknown input type')

    @validator('token')
    def validate_token(cls, token):
        """checks if the token can be load as sensible data
        """
        try:
            if token == 'ask':
                return token
            return get_sensitive_data(token)
        except SensitiveLoadError as e:
            raise ValueError(e)

    @validator('tag')
    def validate_tag(cls, tags):
        """checks if the tag is a list or a string. It always returns a list
        """
        # This validator is implemented to avoid the users to update their
        # tags from string to list.
        # In future, 'tag' will be forced to be a list
        if not isinstance(tags, list):
            logger.warning(
                'Netbox: deprecated string format for tag. Use a list instead')
            tags = [tags]
        return [[t.strip() for t in tag.split(',')] for tag in tags]


class Netbox(Source, InventoryAsyncPlugin):
    """This class is used to dinamically retrieve the inventory from Netbox
       and also retrieve for each device in the inventory its credentials

    """

    def __init__(self, config_data: dict, validate: bool = True) -> None:
        self._status = 'init'
        self._session: aiohttp.ClientSession = None
        self._server: NetboxServerModel = None

        super().__init__(config_data, validate)

    @classmethod
    def get_data_model(cls):
        return NetboxSourceModel

    def _load(self, input_data):
        # load the server class from the dictionary
        if not self._validate:
            input_data['server'] = NetboxServerModel.construct(
                **input_data.pop('url', {}))
            input_data['ssl_verify'] = input_data.pop('ssl-verify', False)
        super()._load(input_data)
        if self._data.token == 'ask':
            self._data.token = get_sensitive_data(
                'ask', f'{self.name} Insert netbox API token: '
            )
        self._server = self._data.server
        if not self._auth:
            raise InventorySourceError(f"{self.name} Netbox must have an "
                                       "'auth' set in the 'namespaces' section"
                                       )

    def _init_session(self, headers: dict):
        """Initialize the session property

        Args:
            headers ([dict]): headers to initialize the session
        """
        if not self._session:
            ssl_option = None if self._data.ssl_verify else False
            self._session = aiohttp.ClientSession(
                headers=headers,
                connector=aiohttp.TCPConnector(ssl=ssl_option)
            )

    def _token_auth_header(self) -> Dict:
        """Generate the token authorization header

        Returns:
            Dict: token authorization header
        """
        return {'Authorization': f'Token {self._data.token}'}

    def _get_url_list(self) -> List[str]:
        """Return the list of requests to execute

        Returns:
            List[str]: list of urls
        """
        urls = []
        url_address = f'{self._server.protocol}://{self._server.host}:'\
            f'{self._server.port}/api/dcim/devices/?'
        for tags in self._data.tag:
            query = url_address
            for (i, t) in enumerate(tags):
                if i > 0:
                    query += '&'
                query += f'tag={t}'
            urls.append(query)
        return urls

    async def get_inventory_list(self) -> List:
        """Contact netbox to retrieve the inventory.

        If more than one tag is set, all devices that have at least one of the
        tags are selected. Netbox api doesn't allow to select with this
        logic by default. If the url query is constructed like
        '?tag=tag1&tag=tag2', only devices with both tags will be returned.
        For this reason, a different request for each tag must be performed.

        Devices with more than one tag may appear duplicated. It's also
        necessary to drop duplicates.

        Raises:
            RuntimeError: Unable to connect to the REST server

        Returns:
            List: inventory list
        """

        if not self._session:
            headers = self._token_auth_header()
            self._init_session(headers)

        # devices is a dictionary to avoid duplicated devices. The key of the
        # dictionary is the device netbox id.
        devices = {}
        try:
            for url in self._get_url_list():
                logger.debug(f"Netbox: Retrieving url '{url}'")
                url_devices, next_url = await self._get_devices(url)
                while next_url:
                    logger.debug(f"Netbox: Retrieving url '{next_url}'")
                    cur_devices, next_url = await self._get_devices(next_url)
                    url_devices.extend(cur_devices)
                devices.update({device['id']: device for device in url_devices
                                if device.get('id') is not None})
        except Exception as e:
            raise InventorySourceError(f'{self.name}: error while '
                                       f'getting devices: {e}')

        logger.info(
            f'Netbox: Retrieved inventory list of {len(devices)} devices')
        return list(devices.values())

    async def _get_devices(self, url: str) -> Tuple[List, str]:
        """Retrieve devices from netbox using an HTTP GET over <url>

        Args:
            url (str): devices url

        Raises:
            RuntimeError: Response error

        Returns:
            Tuple[List, str]: returns the list of devices and the url to get
            the remaining devices
        """
        async with self._session.get(url) as response:
            if int(response.status) == 200:
                res = await response.json()
                next_url = res.get('next')
                if next_url:
                    # The next url might contain a different url if netbox is
                    # behind a reverse proxy.
                    # The code below sets the protocol, the host and the port
                    # of the next request to the same parameters of the
                    # previous request which was successful

                    logger.debug(f'Parsing next page url {next_url}')

                    # parse urls
                    url_data = urlparse(url)
                    next_url_data = urlparse(next_url)

                    # retrieve protocol, host and port from the urls
                    host = url_data.hostname
                    protocol = url_data.scheme or 'http'
                    port = url_data.port or _DEFAULT_PORTS.get(protocol)

                    next_host = next_url_data.hostname
                    next_protocol = next_url_data.scheme or 'http'
                    next_port = (next_url_data.port or
                                 _DEFAULT_PORTS.get(next_protocol))

                    # verify if the two elements are different. If so, log
                    # what's different and set the value to content of the
                    # previous url
                    if host != next_host:
                        logger.debug(
                            'Detected a different host in response: original '
                            f'host "{host}", received host "{next_host}". '
                            f'Setting the request host to "{host}"'
                        )
                        next_host = host
                    if protocol != next_protocol:
                        logger.debug(
                            'Detected a different protocol in response: '
                            f'original protocol "{protocol}", received '
                            f'protocol "{next_protocol}". Setting the request '
                            f'protocol to "{protocol}"'
                        )
                        next_protocol = protocol
                    if port != next_port:
                        logger.debug(
                            'Detected a different port in response: original '
                            f'port "{port}", received port "{next_port}". '
                            f'Setting the request port to "{port}"'
                        )
                        next_port = port

                    # build the next_url
                    next_url = (f'{next_protocol}://{next_host}:{next_port}'
                                f'{next_url_data.path}?{next_url_data.query}')
                return res.get('results', []), next_url
            else:
                raise InventorySourceError(
                    f'{self.name}: error in inventory get '
                    f'{await response.text()}')

    def parse_inventory(self, inventory_list: list) -> Dict:
        """parse the raw inventory collected from the server and generates
           a new inventory with only the required informations

        Args:
            raw_inventory (list): raw inventory received from the server

        Returns:
            List[Dict]: a list containing the inventory
        """
        def get_field_value(entry: dict, fields_str: str):
            # Fields_str is a string containing fields separated by dots
            # Every dot means that the value is deeper in the dictionary
            # Example: fields_str = 'site.name'
            #          return entry[site][name]
            fields = fields_str.split('.')
            cur_field = None
            for i, field in enumerate(fields):
                if i == 0:
                    cur_field = entry.get(field, None)
                else:
                    cur_field = cur_field.get(field, None)
                if cur_field is None:
                    return None
            return cur_field

        inventory = {}
        ignored_device_count = 0

        for device in inventory_list:
            ipv4 = get_field_value(device, 'primary_ip4.address')
            ipv6 = get_field_value(device, 'primary_ip6.address')
            hostname = get_field_value(device, 'name')
            site_name = get_field_value(device, 'site.name')
            if self._namespace == 'netbox-sitename':
                namespace = site_name
            else:
                namespace = self._namespace
            address = None

            if ipv4:
                address = ipv4
            elif ipv6:
                address = ipv6

            if not address:
                logger.warning(
                    f"Skipping {namespace}.{hostname}: doesn't have a "
                    "management IP")
                ignored_device_count += 1
                continue

            address = address.split("/")[0]

            inventory[f'{namespace}.{address}'] = {
                'address': address,
                'namespace': namespace,
                'hostname': hostname,
            }

        logger.info(
            f'Netbox: Acting on inventory of {len(inventory)} devices, '
            f'ignoring {ignored_device_count} devices')

        return inventory

    async def _execute(self):
        while True:
            inventory_list = await self.get_inventory_list()
            logger.debug(
                f"Received netbox inventory from {self._server.protocol}"
                f"://{self._server.host}:{self._server.port}")
            tmp_inventory = self.parse_inventory(inventory_list)
            # Write the inventory and remove the tmp one
            self.set_inventory(tmp_inventory)

            if self._run_once:
                break

            await asyncio.sleep(self._data.period)

    async def _stop(self):
        if self._session:
            await self._session.close()



import pandas as pd
from pandas.core.dtypes.dtypes import DatetimeTZDtype

from suzieq.sqobjects.basicobj import SqObject
from suzieq.shared.utils import humanize_timestamp


class TablesObj(SqObject):
    '''The object providing access to the virtual table: tables'''

    def __init__(self, **kwargs) -> None:
        # We're passing any table name to get init to work
        super().__init__(table='tables', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'query_str']
        self._unique_def_column = ['table']

    def humanize_fields(self, df: pd.DataFrame, _=None) -> pd.DataFrame:
        '''Humanize the timestamp and boot time fields'''
        if df.empty:
            return df

        if 'firstTime' in df.columns:
            if not isinstance(df.firstTime.dtype, DatetimeTZDtype):
                df['firstTime'] = humanize_timestamp(
                    df.firstTime.fillna(0),
                    self.cfg.get('analyzer', {}).get('timezone', None))

        if 'lastTime' in df.columns:
            if not isinstance(df.lastTime.dtype, DatetimeTZDtype):
                df['lastTime'] = humanize_timestamp(
                    df.lastTime.fillna(0),
                    self.cfg.get('analyzer', {}).get('timezone', None))

        return super().humanize_fields(df)



# List of normalized error codes used in the poller

from enum import IntEnum

# Ensure you never copy backwards compat values


class SqPollerStatusCode(IntEnum):
    '''SuzieQ Poller Status Codes and their strings'''

    OK = 0, 'OK'
    HTTP_OK = 200, 'OK'
    PERMISSION_DENIED = 1001, 'Permission Denied'
    UNAUTHORIZED = 1002, 'Unauthorized'
    LOGIN_FAIL = 1003, 'Login Failed'
    PRIV_ESC_FAIL = 1004, 'Privilege Escalation Failed'
    NON_FINAL_FAILURES = 1005, 'Dummy to mark boundary, never set'
    CMD_NOT_FOUND = 2006, 'Command Not Found'
    EMPTY_OUTPUT = 2007, 'Output Empty'
    INCOMPLETE_CMD = 2008, 'Command Incomplete'
    TIMEOUT = 2009, 'Command Timeout'
    DISABLED_FEATURE = 2010, 'Feature Disabled'
    CONNECT_FAIL = 2011, 'Connect Failed'
    PARSING_ERROR = 2012, 'Parsing Failed'
    UNSUPPORTED_SVC = 2013, 'Service Not Supported'
    UNSUPPORTED_CMD = 2014, 'Command Not Supported'
    NO_RT_TO_HOST = 2015, 'No Route to Host'
    VAR_NOT_SET = 2016, 'Variable Not Set'
    WAITING_FOR_VAR = 2017, 'Variable Waiting'
    ERR_OUTPUT = 2018, 'Command Failed'
    GEN_CLS_FAIL = 2019, 'Linux Cmd Failed'
    UNREACH_NET = 2020, 'Network Unreachable'
    NO_SVC_DEF = 2021, 'Service Not Defined'
    UNKNOWN_FAILURE = 2022, 'Unknown Failure'
    CONN_REFUSED = 2023, 'Connection Refused'
    UNEXPECTED_DISCONNECT = 2024, 'Unexpected Disconnect'

    # Everything below this is for backward compat
    LINUX_OLD_CMD_NOT_FOUND = 1, 'Command Not Found'
    NXOS_CMD_NOT_FOUND = 16, 'Command Not Found'
    NXOS_YACNF = 20, 'Command Not Found'
    LINUX_CMD_NOT_FOUND = 127, 'Command Not Found'
    HTTP_EMPTY_OUTPUT = 204, 'Empty Output'
    OLD_EMPTY_OUTPUT = 244, 'Empty Output'
    HTTP_INCOMPLETE_CMD = 400, 'Incomplete Command'
    HTTP_LOGIN_FAIL = 401, 'Login Failed'
    HTTP_BAD_CMD = 404, 'Command Not Found'
    HTTP_YAIC = 405, 'Incomplete Command'
    HTTP_TIMEOUT = 408, 'Timeout'
    OLD_FEATURE_DISABLED = 418, 'Feature Disabled'
    OLD_PARSING_ERROR = 502, 'Parsing Error'
    OLD_PRIV_ESC_FAIL = -1, 'Privilege Escalation Failed'

    def __new__(cls, code: int, errstr: str):
        obj = int.__new__(cls, code)
        obj._value_ = code
        obj.errstr = errstr

        return obj

    def is_error_code_final(self):
        '''is the node in stopped state'''
        return (self.value < self.NON_FINAL_FAILURES.value)



"""
This module contains all the common logic of the
Suzieq poller inventory sources
"""
import abc
import asyncio
import logging
from collections import defaultdict
from contextlib import asynccontextmanager
from typing import Callable, Coroutine, Dict, List

from suzieq.poller.worker.nodes.node import Node
from suzieq.shared.exceptions import SqPollerConfError
from suzieq.shared.sq_plugin import SqPlugin

logger = logging.getLogger(__name__)


class CommandPacer:
    """In many networks, backend authentication servers such as TACACS which
    handle authentication of logins and even command execution, cannot
    large volumes of authentication requests. Thanks to our use of
    asyncio, we can easily sends hundreds of connection requests to such
    servers, which effectively turns into authentication failures. To
    handle this, we add a user-specified maximum of rate of cmds/sec
    that the authentication can handle, and we pace it out. This code
    implements that pacer.
    """
    def __init__(self, max_cmds: int):
        self._max_cmds = max_cmds
        if max_cmds > 0:
            self._cmd_semaphore = asyncio.Semaphore(max_cmds)
            self._cmd_mutex = asyncio.Lock()
            self._cmd_pacer_sleep = float(1 / self.max_cmds)
        else:
            self._cmd_semaphore = None
            self._cmd_mutex = None
            self._cmd_pacer_sleep = 0

    @property
    def max_cmds(self) -> int:
        """Get the maximum number of commands the worker issue every second.
        If there is no limit the returned value is 0.

        Returns:
            int: maximum number of commands
        """
        return self._max_cmds

    @asynccontextmanager
    async def wait(self, use_pacer: bool = True):
        """Context Manager to implement throttling of commands.

        Some networks communicate with a backend authentication server only
        on login while others contact it for authorization of a command as
        well. Its to handle this difference that we pass use_sem. Users set
        the per_cmd_auth to True if authorization is used. The caller of this
        function sets the use_sem apppropriately depending on when the context
        is invoked.

        Args:
            use_pacer(bool): True if you want to use the pacer
        """
        if use_pacer and self._max_cmds:
            async with self._cmd_semaphore:
                async with self._cmd_mutex:
                    await asyncio.sleep(self._cmd_pacer_sleep)
                yield
        else:
            yield


class Inventory(SqPlugin):
    """Inventory is the base class implemented by the
    inventory sources, providing the list of devices
    to poll.
    """

    def __init__(self, add_task_fn: Callable, **kwargs) -> None:
        """Instantiate the Inventory class

        Args:
            add_task_fn (Callable): the function to call to schedule
                a task in the poller.
        """
        self._nodes = {}
        self._node_tasks = {}
        self.add_task_fn = add_task_fn
        self._max_outstanding_cmd = 0
        self._cmd_pacer = None

        self.connect_timeout = kwargs.pop('connect_timeout', 15)
        self.ssh_config_file = kwargs.pop('ssh_config_file', None)

    @property
    def nodes(self) -> Dict[str, Node]:
        """Get the current list of nodes in the inventory

        Returns:
            Dict[str, Node]: get a dictionary with the nodes in the inventory
                with format {namespace.hostname: Node}
        """
        return self._nodes

    @property
    def running_nodes(self) -> List[Coroutine]:
        """Get the the couroutines of the running nodes

        Returns:
            List[Coroutine]: a list containing the coroutines of the running
                nodes
        """
        return self._node_tasks

    async def build_inventory(self) -> Dict[str, Node]:
        """Retrieve the list of nodes to poll and instantiate
        all the Nodes objects in the retrieved inventory.

        Raises:
            SqPollerConfError: in case of wrong inventory configuration
            InventorySourceError: in case of error with the inventory source

        Returns:
            Dict[str, Node]: a list containing all the nodes in the inventory
        """

        inventory_list = await self._get_device_list()
        if not inventory_list:
            raise SqPollerConfError('The inventory source returned no hosts')

        self._cmd_pacer = CommandPacer(self._max_outstanding_cmd)

        # Initialize the nodes in the inventory
        self._nodes = await self._init_nodes(inventory_list)
        return self._nodes

    def get_node_callq(self) -> Dict[str, Dict]:
        """Get the dictionary allowing to send command query
        on the nodes

        Returns:
            Dict[str, Dict]: a dictionary having, for each 'namespace.hostname'
                the node hostname and the function allowing to call
                a query on the node.
        """
        node_callq = defaultdict(lambda: defaultdict(dict))

        node_callq.update({x: {'hostname': self._nodes[x].hostname,
                               'postq':    self._nodes[x].post_commands}
                           for x in self._nodes})

        return node_callq

    async def schedule_nodes_run(self):
        """Schedule the nodes tasks in the poller, so that they can
        start processing the command query passed to the nodes.

        This function should be called only once, if called more
        than once, it won't have any effect.
        """
        if not self._node_tasks:
            self._node_tasks = {node: self._nodes[node].run()
                                for node in self._nodes}
            await self.add_task_fn(list(self._node_tasks.values()))

    async def _init_nodes(self, inventory_list:
                          List[Dict]) -> Dict[str, Node]:
        """Initialize the Node objects given the of credentials of the nodes.
        After this function is called, a connection with the nodes in the list
        is performed.

        Returns:
            Dict[str, Node]: a list containing the initialized Node objects
        """
        init_tasks = []
        nodes_list = {}

        for host in inventory_list:
            new_node = Node()
            init_tasks += [new_node.initialize(
                **host,
                cmd_pacer=self._cmd_pacer,
                connect_timeout=self.connect_timeout,
                ssh_config_file=self.ssh_config_file
            )]

        for n in asyncio.as_completed(init_tasks):
            try:
                newnode = await n
            except Exception as e:  # pylint: disable=broad-except
                logger.error(
                    f'Encountered error {str(e)} in initializing node')
                continue
            if newnode.devtype == "unsupported":
                logger.error(
                    f'Unsupported device type for '
                    f'{newnode.address}.{newnode.port}'
                )
            elif not newnode.devtype:
                logger.warning('Unable to determine device type for '
                               f'{newnode.address}. Will retry')
            else:
                logger.info(f'Added node {newnode.hostname}:{newnode.port} '
                            f'of type {newnode.devtype}')
            nodes_list.update({self.get_node_key(newnode): newnode})

        return nodes_list

    @ abc.abstractmethod
    async def _get_device_list(self) -> List[Dict]:
        """Retrieve the devices credentials from the inventory
        source

        Raises:
            InventorySourceError: in case of error with the inventory source

        Returns:
            List[Dict]: the list of the credentials of the devices
                in the Suzieq native inventory file.
        """
        raise NotImplementedError

    @ staticmethod
    def get_node_key(node: Node) -> str:
        """Given a node object it returns its ID key

        Args:
            node (Node): the node from which retrieving the key

        Returns:
            str: a string containing the ID key of the node
        """
        return f'{node.nsname}.{node.hostname}'



from typing import List
from suzieq.sqobjects.basicobj import SqObject


class EvpnvniObj(SqObject):
    '''The object providing access to the evpnVni table'''

    def __init__(self, **kwargs):
        super().__init__(table='evpnVni', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'vni',
                                'priVtepIp', 'query_str']
        self._valid_assert_args = self._valid_get_args + ['result']
        self._valid_arg_vals = {
            'result': ['all', 'pass', 'fail'],
        }

    def _get_empty_cols(self, columns: List[str], fun: str, **kwargs) \
            -> List[str]:
        if fun == 'assert':
            if columns in [['default'], ['*']]:
                # this is the copy of engine/pandas/self._assert_result_cols
                return ['namespace', 'hostname', 'vni', 'type', 'vrf',
                        'macaddr', 'timestamp', 'result', 'assertReason']
        return super()._get_empty_cols(columns, fun, **kwargs)
